package capacity

import "testing"

func required(req int, info CapacityInfo) func(t *testing.T) {
	return func(t *testing.T) {
		t.Helper()

		n := RequiredInstances(info)
		if n != req {
			t.Errorf("expected %d, got %d", req, n)
		}
	}
}

func TestCapacityBasic(t *testing.T) {
	t.Run("1 acquired, 1 capacity per instance, needs 1 instance", required(1, CapacityInfo{
		Acquired:            1,
		CapacityPerInstance: 1,
	}))

	t.Run("100 acquired, 100 capacity per instance, needs 1 instance", required(1, CapacityInfo{
		Acquired:            100,
		CapacityPerInstance: 100,
	}))

	t.Run("1 pending, 1 capacity per instance, needs 1 instance", required(1, CapacityInfo{
		Pending:             1,
		CapacityPerInstance: 1,
	}))

	t.Run("100 pending, 100 capacity per instance, needs 1 instance", required(1, CapacityInfo{
		Pending:             100,
		CapacityPerInstance: 100,
	}))

	t.Run("100 acquired, 100 pending, 100 capacity per instance, needs 2 instance", required(2, CapacityInfo{
		Acquired:            100,
		Pending:             100,
		CapacityPerInstance: 100,
	}))

	t.Run("100 acquired, 100 pending, 100 capacity per instance, 1 max instance count, needs 1 instance", required(1, CapacityInfo{
		Acquired:            100,
		Pending:             100,
		CapacityPerInstance: 100,
		MaxInstanceCount:    1,
	}))
}

func TestCapacity2xScaleFactor(t *testing.T) {
	t.Run("1 acquired, 1 capacity per instance, needs 2 instances", required(2, CapacityInfo{
		Acquired:            1,
		ScaleFactor:         2,
		CapacityPerInstance: 1,
	}))

	t.Run("100 acquired, 100 capacity per instance, needs 2 instances", required(2, CapacityInfo{
		Acquired:            100,
		ScaleFactor:         2,
		CapacityPerInstance: 100,
	}))

	t.Run("1 pending, 1 capacity per instance, needs 1 instance", required(1, CapacityInfo{
		Pending:             1,
		ScaleFactor:         2,
		CapacityPerInstance: 1,
	}))

	t.Run("100 pending, 100 capacity per instance, needs 1 instance", required(1, CapacityInfo{
		Pending:             100,
		ScaleFactor:         2,
		CapacityPerInstance: 100,
	}))

	t.Run("100 acquired, 100 pending, 100 capacity per instance, needs 2 instance", required(2, CapacityInfo{
		Acquired:            100,
		Pending:             100,
		ScaleFactor:         2,
		CapacityPerInstance: 100,
	}))

	t.Run("100 acquired, 100 pending, 100 capacity per instance, 2 max instance count, needs 2 instance", required(2, CapacityInfo{
		Acquired:            100,
		Pending:             100,
		ScaleFactor:         2,
		CapacityPerInstance: 100,
		MaxInstanceCount:    2,
	}))

	t.Run("100 acquired, 100 pending, 100 capacity per instance, 1 max instance count, needs 1 instance", required(1, CapacityInfo{
		Acquired:            100,
		Pending:             100,
		ScaleFactor:         2,
		CapacityPerInstance: 100,
		MaxInstanceCount:    1,
	}))
}

func TestCapacityUsedCounts(t *testing.T) {
	t.Run("1 instance, 2 capacity per instance, one unavailable, nothing pending", required(0, CapacityInfo{
		InstanceCount:       1,
		Acquired:            1,
		CapacityPerInstance: 2,
		UnavailableCapacity: 1,
	}))

	t.Run("1 instance, 2 capacity per instance, one unavailable, one pending", required(1, CapacityInfo{
		InstanceCount:       1,
		Acquired:            1,
		CapacityPerInstance: 2,
		Pending:             1,
		UnavailableCapacity: 1,
	}))

	t.Run("2 instances, 2 capacity per instance, two unavailable, two pending", required(0, CapacityInfo{
		InstanceCount:       2,
		Acquired:            0,
		CapacityPerInstance: 2,
		Pending:             2,
		UnavailableCapacity: 2,
	}))

	t.Run("2 instances, 2 capacity per instance, two unavailable, two pending", required(0, CapacityInfo{
		InstanceCount:       2,
		Acquired:            0,
		CapacityPerInstance: 2,
		Pending:             2,
		UnavailableCapacity: 2,
	}))
}

func TestCapacityOnePerInstance(t *testing.T) {
	t.Run("at capacity, nothing pending", required(0, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
	}))

	t.Run("at capacity, 1 pending", required(1, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             1,
	}))

	t.Run("at capacity, 5 pending", required(5, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             5,
	}))

	t.Run("at capacity, 1 pending, 1 idle", required(1, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             1,
		IdleCount:           1,
	}))

	t.Run("at capacity, 1 pending, 5 idle", required(5, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             1,
		IdleCount:           5,
	}))

	t.Run("at capacity, 5 pending, 5 idle", required(5, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           5,
	}))

	t.Run("at capacity, 5 pending, 1 idle", required(5, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           1,
	}))

	t.Run("at capacity, 5 pending, 1 idle, 10 max instance count", required(0, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    10,
	}))

	t.Run("at capacity, 5 pending, 1 idle, 5 max instance count", required(0, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    5,
	}))

	t.Run("at capacity, 5 pending, 1 idle, 4 max instance count", required(0, CapacityInfo{
		InstanceCount:       10,
		Acquired:            10,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    4,
	}))

	t.Run("under capacity, nothing pending, 1 idle", required(0, CapacityInfo{
		InstanceCount:       10,
		Acquired:            9,
		CapacityPerInstance: 1,
		IdleCount:           1,
	}))

	t.Run("under capacity, nothing pending", required(-1, CapacityInfo{
		InstanceCount:       10,
		Acquired:            9,
		CapacityPerInstance: 1,
	}))

	t.Run("under capacity, 1 pending", required(-10, CapacityInfo{
		InstanceCount:       20,
		Acquired:            9,
		CapacityPerInstance: 1,
		Pending:             1,
	}))

	t.Run("under capacity, 5 pending, 1 idle", required(-6, CapacityInfo{
		InstanceCount:       20,
		Acquired:            9,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           1,
	}))

	t.Run("under capacity, 5 pending, 1 idle, 5 max instance count", required(-6, CapacityInfo{
		InstanceCount:       20,
		Acquired:            9,
		CapacityPerInstance: 1,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    5,
	}))
}

func TestCapacityTwoPerInstance(t *testing.T) {
	t.Run("at capacity, nothing pending", required(0, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
	}))

	t.Run("at capacity, 1 pending", required(1, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             1,
	}))

	t.Run("at capacity, 5 pending", required(3, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             5,
	}))

	t.Run("at capacity, 1 pending, 1 idle", required(1, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             1,
		IdleCount:           1,
	}))

	t.Run("at capacity, 1 pending, 5 idle", required(3, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             1,
		IdleCount:           5,
	}))

	t.Run("at capacity, 5 pending, 5 idle", required(3, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           5,
	}))

	t.Run("at capacity, 5 pending, 1 idle", required(3, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           1,
	}))

	t.Run("at capacity, 5 pending, 1 idle, 10 max instance count", required(3, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    10,
	}))

	t.Run("at capacity, 5 pending, 1 idle, 3 max instance count", required(0, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    3,
	}))

	t.Run("at capacity, 5 pending, 1 idle, 2 max instance count", required(0, CapacityInfo{
		InstanceCount:       5,
		Acquired:            10,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    2,
	}))

	t.Run("under capacity, nothing pending, 1 idle", required(0, CapacityInfo{
		InstanceCount:       5,
		Acquired:            9,
		CapacityPerInstance: 2,
		IdleCount:           1,
	}))

	t.Run("under capacity, nothing pending, 2 idle", required(1, CapacityInfo{
		InstanceCount:       5,
		Acquired:            9,
		CapacityPerInstance: 2,
		IdleCount:           2,
	}))

	t.Run("under capacity, nothing pending", required(-1, CapacityInfo{
		InstanceCount:       5,
		Acquired:            8,
		CapacityPerInstance: 2,
	}))

	t.Run("under capacity, 1 pending", required(-5, CapacityInfo{
		InstanceCount:       10,
		Acquired:            9,
		CapacityPerInstance: 2,
		Pending:             1,
	}))

	t.Run("under capacity, 5 pending, 1 idle", required(-3, CapacityInfo{
		InstanceCount:       10,
		Acquired:            9,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           1,
	}))

	t.Run("under capacity, 5 pending, 1 idle, 10 max instance count", required(-3, CapacityInfo{
		InstanceCount:       10,
		Acquired:            9,
		CapacityPerInstance: 2,
		Pending:             5,
		IdleCount:           1,
		MaxInstanceCount:    10,
	}))
}

func TestCapacityReservations(t *testing.T) {
	t.Run("1 acquired, 1 capacity per instance, 1 reserved, needs 1 instance", required(1, CapacityInfo{
		Acquired:            1,
		CapacityPerInstance: 1,
		Reserved:            1,
	}))

	t.Run("10 instances, 2 acquired, 1 capacity per instance, 5 idle, 0 reserved, needs -3", required(-3, CapacityInfo{
		InstanceCount:       10,
		Acquired:            2,
		CapacityPerInstance: 1,
		IdleCount:           5,
		Reserved:            0,
	}))

	t.Run("10 instances, 2 acquired, 1 capacity per instance, 5 idle, 2 reserved, needs -1", required(-1, CapacityInfo{
		InstanceCount:       10,
		Acquired:            2,
		CapacityPerInstance: 1,
		IdleCount:           5,
		Reserved:            2,
	}))

	t.Run("10 instances, 2 acquired, 5 capacity per instance, 5 idle, 2 reserved, needs -8", required(-8, CapacityInfo{
		InstanceCount:       10,
		Acquired:            2,
		CapacityPerInstance: 5,
		IdleCount:           5,
		Reserved:            2,
	}))

	t.Run("10 instances, 2 acquired, 5 capacity per instance, 5 idle, 5 reserved, needs -7", required(-7, CapacityInfo{
		InstanceCount:       10,
		Acquired:            2,
		CapacityPerInstance: 5,
		IdleCount:           5,
		Reserved:            5,
	}))

	t.Run("10 instances, 2 acquired, 5 capacity per instance, 5 idle, 5 reserved, 10 max instance count, needs -7", required(-7, CapacityInfo{
		InstanceCount:       10,
		Acquired:            2,
		CapacityPerInstance: 5,
		IdleCount:           5,
		Reserved:            5,
		MaxInstanceCount:    10,
	}))
}
